/*
 * Copyright 2014 Google Inc. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.inferred.freebuilder.processor.property;

import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.annotation.JsonDeserialize;
import com.fasterxml.jackson.datatype.jdk8.Jdk8Module;
import com.google.common.collect.Lists;
import com.google.common.primitives.Primitives;
import com.google.common.testing.EqualsTester;

import org.inferred.freebuilder.FreeBuilder;
import org.inferred.freebuilder.processor.FeatureSets;
import org.inferred.freebuilder.processor.NamingConvention;
import org.inferred.freebuilder.processor.Processor;
import org.inferred.freebuilder.processor.source.SourceBuilder;
import org.inferred.freebuilder.processor.source.feature.FeatureSet;
import org.inferred.freebuilder.processor.source.testing.BehaviorTester;
import org.inferred.freebuilder.processor.source.testing.ParameterizedBehaviorTestFactory;
import org.inferred.freebuilder.processor.source.testing.ParameterizedBehaviorTestFactory.Shared;
import org.inferred.freebuilder.processor.source.testing.TestBuilder;
import org.junit.Rule;
import org.junit.Test;
import org.junit.rules.ExpectedException;
import org.junit.runner.RunWith;
import org.junit.runners.Parameterized;
import org.junit.runners.Parameterized.Parameters;
import org.junit.runners.Parameterized.UseParametersRunnerFactory;

import java.util.Arrays;
import java.util.List;
import java.util.OptionalDouble;
import java.util.OptionalInt;
import java.util.OptionalLong;
import java.util.function.IntFunction;

/** Behavioral tests for primitive optional properties. */
@RunWith(Parameterized.class)
@UseParametersRunnerFactory(ParameterizedBehaviorTestFactory.class)
public class PrimitiveOptionalPropertyTest {

  enum OptionalFactory {
    INTS(OptionalInt.class, int.class, v -> v * (v + 1) / 2),
    LONGS(OptionalLong.class, long.class, v -> (long) v * (v + 1) / 2),
    DOUBLE(OptionalDouble.class, double.class, v -> 1.0 / (v + 1));

    private final Class<?> type;
    private final Class<? extends Number> primitiveType;
    private final Class<?> boxedType;
    private final IntFunction<? extends Number> examples;

    <T, P extends Number> OptionalFactory(
        Class<T> type,
        Class<P> primitiveType,
        IntFunction<P> examples) {
      this.type = type;
      this.primitiveType = primitiveType;
      this.boxedType = Primitives.wrap(primitiveType);
      this.examples = examples;
    }

    Number example(int i) {
      return examples.apply(i);
    }

    @Override
    public String toString() {
      return type.getSimpleName();
    }
  }

  private static final String INVALID_MESSAGE = "item must be non-negative";

  @SuppressWarnings("unchecked")
  @Parameters(name = "{0}, {1}, {2}")
  public static Iterable<Object[]> parameters() {
    List<OptionalFactory> optionals = Arrays.asList(OptionalFactory.values());
    List<NamingConvention> conventions = Arrays.asList(NamingConvention.values());
    List<FeatureSet> features = FeatureSets.ALL;
    return () -> Lists
        .cartesianProduct(optionals, conventions, features)
        .stream()
        .map(List::toArray)
        .iterator();
  }

  @Rule public final ExpectedException thrown = ExpectedException.none();
  @Shared public BehaviorTester behaviorTester;

  private final OptionalFactory optional;
  private final NamingConvention convention;
  private final FeatureSet features;

  private final SourceBuilder datatype;

  public PrimitiveOptionalPropertyTest(
      OptionalFactory optional,
      NamingConvention convention,
      FeatureSet features) {
    this.optional = optional;
    this.convention = convention;
    this.features = features;

    datatype = SourceBuilder.forTesting()
        .addLine("package com.example;")
        .addLine("@%s", FreeBuilder.class)
        .addLine("@%s(builder = DataType.Builder.class)", JsonDeserialize.class)
        .addLine("public interface DataType {")
        .addLine("  %s %s;", optional.type, convention.get("item"))
        .addLine("")
        .addLine("  class Builder extends DataType_Builder {")
        .addLine("    @Override public Builder %s(%s item) {",
            convention.set("item"), optional.primitiveType)
        .addLine("      if (item < 0) {")
        .addLine("        throw new IllegalArgumentException(\"%s\");", INVALID_MESSAGE)
        .addLine("      }")
        .addLine("      return super.%s(item);", convention.set("item"))
        .addLine("    }")
        .addLine("  }")
        .addLine("")
        .addLine("  public static Builder builder() {")
        .addLine("    return new Builder();")
        .addLine("  }")
        .addLine("}");
  }

  @Test
  public void testConstructor_defaultEmpty() {
    behaviorTester
        .with(new Processor(features))
        .with(datatype)
        .with(testBuilder()
            .addLine("DataType value = new DataType.Builder().build();")
            .addLine("assertEquals(%s.empty(), value.%s);", optional.type, convention.get("item"))
            .build())
        .runTest();
  }

  @Test
  public void testBuilderGetter_defaultValue() {
    behaviorTester
        .with(new Processor(features))
        .with(datatype)
        .with(testBuilder()
            .addLine("DataType.Builder builder = new DataType.Builder();")
            .addLine("assertEquals(%s.empty(), builder.%s);", optional.type, convention.get("item"))
            .build())
        .runTest();
  }

  @Test
  public void testBuilderGetter_nonDefaultValue() {
    behaviorTester
        .with(new Processor(features))
        .with(datatype)
        .with(testBuilder()
            .addLine("DataType.Builder builder = new DataType.Builder()")
            .addLine("    .%s(%s);", convention.set("item"), optional.example(0))
            .addLine("assertEquals(%s.of(%s), builder.%s);",
                optional.type, optional.example(0), convention.get("item"))
            .build())
        .runTest();
  }

  @Test
  public void testSet_primitive() {
    behaviorTester
        .with(new Processor(features))
        .with(datatype)
        .with(testBuilder()
            .addLine("DataType value = new DataType.Builder()")
            .addLine("    .%s(%s)", convention.set("item"), optional.example(0))
            .addLine("    .build();")
            .addLine("assertEquals(%s.of(%s), value.%s);",
                optional.type, optional.example(0), convention.get("item"))
            .build())
        .runTest();
  }

  @Test
  public void testSet_optional() {
    behaviorTester
        .with(new Processor(features))
        .with(datatype)
        .with(testBuilder()
            .addLine("DataType value = new DataType.Builder()")
            .addLine("    .%s(%s.of(%s))", convention.set("item"),
                optional.type, optional.example(0))
            .addLine("    .build();")
            .addLine("assertEquals(%s.of(%s), value.%s);",
                optional.type, optional.example(0), convention.get("item"))
            .build())
        .runTest();
  }

  @Test
  public void testSet_optionalDelegatesToPrimitiveSetterForValidation() {
    thrown.expect(IllegalArgumentException.class);
    thrown.expectMessage(INVALID_MESSAGE);
    behaviorTester
        .with(new Processor(features))
        .with(datatype)
        .with(testBuilder()
            .addLine("DataType.Builder template = DataType.builder()")
            .addLine("    .%s(%s.of(-1));", convention.set("item"), optional.type)
            .build())
        .runTest();
  }

  @Test
  public void testSet_empty() {
    behaviorTester
        .with(new Processor(features))
        .with(datatype)
        .with(testBuilder()
            .addLine("DataType value = new DataType.Builder()")
            .addLine("    .%s(%s.empty())", convention.set("item"), optional.type)
            .addLine("    .build();")
            .addLine("assertEquals(%s.empty(), value.%s);", optional.type, convention.get("item"))
            .build())
        .runTest();
  }

  @Test
  public void testSet_nullOptional() {
    thrown.expect(NullPointerException.class);
    behaviorTester
        .with(new Processor(features))
        .with(datatype)
        .with(testBuilder()
            .addLine("new DataType.Builder().%s((%s) null);", convention.set("item"), optional.type)
            .build())
        .runTest();
  }

  @Test
  public void testMap_modifiesValue() {
    behaviorTester
        .with(new Processor(features))
        .with(datatype)
        .with(testBuilder()
            .addLine("DataType value = new DataType.Builder()")
            .addLine("    .%s(%s)", convention.set("item"), optional.example(1))
            .addLine("    .mapItem(a -> a + %s)", optional.example(2))
            .addLine("    .build();")
            .addLine("assertEquals(%s.of(%s + %s), value.%s);",
                optional.type, optional.example(1), optional.example(2), convention.get("item"))
            .build())
        .runTest();
  }

  @Test
  public void testMap_delegatesToSetterForValidation() {
    thrown.expect(IllegalArgumentException.class);
    thrown.expectMessage(INVALID_MESSAGE);
    behaviorTester
        .with(new Processor(features))
        .with(datatype)
        .with(testBuilder()
            .addLine("new DataType.Builder()")
            .addLine("    .%s(%s)", convention.set("item"), optional.example(0))
            .addLine("    .mapItem(a -> -1);")
            .build())
        .runTest();
  }

  @Test
  public void testMap_throwsNpeIfMapperIsNull() {
    thrown.expect(NullPointerException.class);
    behaviorTester
        .with(new Processor(features))
        .with(datatype)
        .with(testBuilder()
            .addLine("new DataType.Builder()")
            .addLine("    .%s(%s)", convention.set("item"), optional.example(0))
            .addLine("    .mapItem(null);")
            .build())
        .runTest();
  }

  @Test
  public void testMap_throwsNpeIfMapperIsNullAndPropertyIsEmpty() {
    thrown.expect(NullPointerException.class);
    behaviorTester
        .with(new Processor(features))
        .with(datatype)
        .with(testBuilder()
            .addLine("new DataType.Builder()")
            .addLine("    .mapItem(null);")
            .build())
        .runTest();
  }

  @Test
  public void testMap_skipsMapperIfPropertyIsEmpty() {
    behaviorTester
        .with(new Processor(features))
        .with(datatype)
        .with(testBuilder()
            .addLine("DataType value = new DataType.Builder()")
            .addLine("    .mapItem(a -> { fail(\"mapper called\"); return -1; })")
            .addLine("    .build();")
            .addLine("assertFalse(value.%s.isPresent());", convention.get("item"))
            .build())
        .runTest();
  }

  @Test
  public void testMap_canUseCustomBoxedFunctionalInterface() {
    SourceBuilder customMutatorType = SourceBuilder.forTesting();
    for (String line : datatype.toString().split("\n")) {
      if (line.contains("extends DataType_Builder")) {
        int insertOffset = line.indexOf('{') + 1;
        customMutatorType
            .addLine("%s", line.substring(0, insertOffset))
            .addLine("    public interface Mapper {")
            .addLine("      %1$s map(%1$s value);", optional.boxedType)
            .addLine("    }")
            .addLine("    @Override public Builder mapItem(Mapper mapper) {")
            .addLine("      return super.mapItem(mapper);")
            .addLine("    }")
            .addLine("%s", line.substring(insertOffset));
      } else {
        customMutatorType.addLine("%s", line);
      }
    }

    behaviorTester
        .with(new Processor(features))
        .with(customMutatorType)
        .with(testBuilder()
            .addLine("DataType value = new DataType.Builder()")
            .addLine("    .%s(%s)", convention.set("item"), optional.example(0))
            .addLine("    .mapItem(a -> null)")
            .addLine("    .build();")
            .addLine("assertEquals(%s.empty(), value.%s);", optional.type, convention.get("item"))
            .build())
        .runTest();
  }

  @Test
  public void testMap_canUseCustomUnboxedFunctionalInterface() {
    SourceBuilder customMutatorType = SourceBuilder.forTesting();
    for (String line : datatype.toString().split("\n")) {
      if (line.contains("extends DataType_Builder")) {
        int insertOffset = line.indexOf('{') + 1;
        customMutatorType
            .addLine("%s", line.substring(0, insertOffset))
            .addLine("    public interface Mapper {")
            .addLine("      %1$s map(%1$s value);", optional.primitiveType)
            .addLine("    }")
            .addLine("    @Override public Builder mapItem(Mapper mapper) {")
            .addLine("      return super.mapItem(mapper);")
            .addLine("    }")
            .addLine("%s", line.substring(insertOffset));
      } else {
        customMutatorType.addLine("%s", line);
      }
    }

    behaviorTester
        .with(new Processor(features))
        .with(customMutatorType)
        .with(testBuilder()
            .addLine("DataType value = new DataType.Builder()")
            .addLine("    .%s(%s)", convention.set("item"), optional.example(0))
            .addLine("    .mapItem(a -> %s)", optional.example(1))
            .addLine("    .build();")
            .addLine("assertEquals(%s.of(%s), value.%s);",
                optional.type, optional.example(1), convention.get("item"))
            .build())
        .runTest();
  }

  @Test
  public void testMap_canUseCustomOptionalFunctionalInterface() {
    SourceBuilder customMutatorType = SourceBuilder.forTesting();
    for (String line : datatype.toString().split("\n")) {
      if (line.contains("extends DataType_Builder")) {
        int insertOffset = line.indexOf('{') + 1;
        customMutatorType
            .addLine("%s", line.substring(0, insertOffset))
            .addLine("    public interface Mapper {")
            .addLine("      %1$s map(%1$s value);", optional.type)
            .addLine("    }")
            .addLine("    @Override public Builder mapItem(Mapper mapper) {")
            .addLine("      return super.mapItem(mapper);")
            .addLine("    }")
            .addLine("%s", line.substring(insertOffset));
      } else {
        customMutatorType.addLine("%s", line);
      }
    }

    behaviorTester
        .with(new Processor(features))
        .with(customMutatorType)
        .with(testBuilder()
            .addLine("DataType value = new DataType.Builder()")
            .addLine("    .%s(%s)", convention.set("item"), optional.example(0))
            .addLine("    .mapItem(a -> %s.empty())", optional.type)
            .addLine("    .build();")
            .addLine("assertEquals(%s.empty(), value.%s);", optional.type, convention.get("item"))
            .build())
        .runTest();
  }

  @Test
  public void testClear() {
    behaviorTester
        .with(new Processor(features))
        .with(datatype)
        .with(testBuilder()
            .addLine("DataType value = new DataType.Builder()")
            .addLine("    .%s(%s)", convention.set("item"), optional.example(0))
            .addLine("    .clearItem()")
            .addLine("    .build();")
            .addLine("assertEquals(%s.empty(), value.%s);", optional.type, convention.get("item"))
            .build())
        .runTest();
  }

  @Test
  public void testMergeFrom_valueInstance() {
    behaviorTester
        .with(new Processor(features))
        .with(datatype)
        .with(testBuilder()
            .addLine("DataType value = DataType.builder()")
            .addLine("    .%s(%s)", convention.set("item"), optional.example(0))
            .addLine("    .build();")
            .addLine("DataType.Builder builder = DataType.builder()")
            .addLine("    .mergeFrom(value);")
            .addLine("assertEquals(%s.of(%s), builder.%s);",
                optional.type, optional.example(0), convention.get("item"))
            .build())
        .runTest();
  }

  @Test
  public void testMergeFrom_builder() {
    behaviorTester
        .with(new Processor(features))
        .with(datatype)
        .with(testBuilder()
            .addLine("DataType.Builder template = DataType.builder()")
            .addLine("    .%s(%s);", convention.set("item"), optional.example(0))
            .addLine("DataType.Builder builder = DataType.builder()")
            .addLine("    .mergeFrom(template);")
            .addLine("assertEquals(%s.of(%s), builder.%s);",
                optional.type, optional.example(0), convention.get("item"))
            .build())
        .runTest();
  }

  @Test
  public void testMergeFrom_valueInstance_emptyOptional() {
    behaviorTester
        .with(new Processor(features))
        .with(datatype)
        .with(testBuilder()
            .addLine("DataType value = DataType.builder()")
            .addLine("    .build();")
            .addLine("DataType.Builder builder = DataType.builder()")
            .addLine("    .%s(%s)", convention.set("item"), optional.example(0))
            .addLine("    .mergeFrom(value);")
            .addLine("assertEquals(%s.of(%s), builder.%s);",
                optional.type, optional.example(0), convention.get("item"))
            .build())
        .runTest();
  }

  @Test
  public void testMergeFrom_builder_emptyOptional() {
    behaviorTester
        .with(new Processor(features))
        .with(datatype)
        .with(testBuilder()
            .addLine("DataType.Builder template = DataType.builder();")
            .addLine("DataType.Builder builder = DataType.builder()")
            .addLine("    .%s(%s)", convention.set("item"), optional.example(0))
            .addLine("    .mergeFrom(template);")
            .addLine("assertEquals(%s.of(%s), builder.%s);",
                optional.type, optional.example(0), convention.get("item"))
            .build())
        .runTest();
  }

  @Test
  public void testBuilderClear() {
    behaviorTester
        .with(new Processor(features))
        .with(datatype)
        .with(testBuilder()
            .addLine("DataType value = new DataType.Builder()")
            .addLine("    .%s(%s)", convention.set("item"), optional.example(0))
            .addLine("    .clear()")
            .addLine("    .build();")
            .addLine("assertEquals(%s.empty(), value.%s);", optional.type, convention.get("item"))
            .build())
        .runTest();
  }

  @Test
  public void testBuilderClear_customDefault() {
    behaviorTester
        .with(new Processor(features))
        .with(SourceBuilder.forTesting()
            .addLine("package com.example;")
            .addLine("@%s", FreeBuilder.class)
            .addLine("public abstract class DataType {")
            .addLine("  public abstract %s %s;", optional.type, convention.get("item"))
            .addLine("")
            .addLine("  public static class Builder extends DataType_Builder {}")
            .addLine("  public static Builder builder() {")
            .addLine("    return new Builder().%s(%s);",
                convention.set("item"), optional.example(1))
            .addLine("  }")
            .addLine("}"))
        .with(testBuilder()
            .addLine("DataType value = DataType.builder()")
            .addLine("    .%s(%s)", convention.set("item"), optional.example(0))
            .addLine("    .clear()")
            .addLine("    .build();")
            .addLine("assertEquals(%s.of(%s), value.%s);",
                optional.type, optional.example(1), convention.get("item"))
            .build())
        .runTest();
  }

  @Test
  public void testBuilderClear_noBuilderFactory() {
    behaviorTester
        .with(new Processor(features))
        .with(SourceBuilder.forTesting()
            .addLine("package com.example;")
            .addLine("@%s", FreeBuilder.class)
            .addLine("public abstract class DataType {")
            .addLine("  public abstract %s %s;", optional.type, convention.get("item"))
            .addLine("")
            .addLine("  public static class Builder extends DataType_Builder {")
            .addLine("    public Builder(%s s) {", optional.primitiveType)
            .addLine("      %s(s);", convention.set("item"))
            .addLine("    }")
            .addLine("  }")
            .addLine("}"))
        .with(testBuilder()
            .addLine("DataType value = new DataType.Builder(%s)", optional.example(0))
            .addLine("    .clear()")
            .addLine("    .build();")
            .addLine("assertEquals(%s.empty(), value.%s);", optional.type, convention.get("item"))
            .build())
        .runTest();
  }

  @Test
  public void testEquality() {
    behaviorTester
        .with(new Processor(features))
        .with(datatype)
        .with(testBuilder()
            .addLine("new %s()", EqualsTester.class)
            .addLine("    .addEqualityGroup(")
            .addLine("        DataType.builder().build(),")
            .addLine("        DataType.builder()")
            .addLine("            .%s(%s.empty())", convention.set("item"), optional.type)
            .addLine("            .build(),")
            .addLine("        DataType.builder()")
            .addLine("            .%s(%s)", convention.set("item"), optional.example(0))
            .addLine("            .clearItem()")
            .addLine("            .build())")
            .addLine("    .addEqualityGroup(")
            .addLine("        DataType.builder()")
            .addLine("            .%s(%s)", convention.set("item"), optional.example(0))
            .addLine("            .build(),")
            .addLine("        DataType.builder()")
            .addLine("            .%s(%s.of(%s))",
                convention.set("item"), optional.type, optional.example(0))
            .addLine("            .build())")
            .addLine("    .testEquals();")
            .build())
        .runTest();
  }

  @Test
  public void testValueToString() {
    behaviorTester
        .with(new Processor(features))
        .with(datatype)
        .with(testBuilder()
            .addLine("DataType empty = DataType.builder()")
            .addLine("    .build();")
            .addLine("DataType present = DataType.builder()")
            .addLine("    .%s(%s)", convention.set("item"), optional.example(0))
            .addLine("    .build();")
            .addLine("assertEquals(\"DataType{}\", empty.toString());")
            .addLine("assertEquals(\"DataType{item=\" + %s + \"}\", present.toString());",
                optional.example(0))
            .build())
        .runTest();
  }

  @Test
  public void testPartialToString() {
    behaviorTester
        .with(new Processor(features))
        .with(datatype)
        .with(testBuilder()
            .addLine("DataType empty = DataType.builder()")
            .addLine("    .buildPartial();")
            .addLine("DataType present = DataType.builder()")
            .addLine("    .%s(%s)", convention.set("item"), optional.example(0))
            .addLine("    .buildPartial();")
            .addLine("assertEquals(\"partial DataType{}\", empty.toString());")
            .addLine("assertEquals(\"partial DataType{item=\" + %s + \"}\", present.toString());",
                optional.example(0))
            .build())
        .runTest();
  }

  @Test
  public void testJacksonInteroperability() {
    behaviorTester
        .with(new Processor(features))
        .with(datatype)
        .with(testBuilder()
            .addLine("DataType value = new DataType.Builder().%s(%s).build();",
                convention.set("item"), optional.example(0))
            .addLine("%1$s mapper = new %1$s()", ObjectMapper.class)
            .addLine("    .registerModule(new %s());", Jdk8Module.class)
            .addLine("String json = mapper.writeValueAsString(value);")
            .addLine("assertEquals(\"{\\\"item\\\":%s}\", json);", optional.example(0))
            .addLine("DataType clone = mapper.readValue(json, DataType.class);")
            .addLine("assertEquals(%s.of(%s), clone.%s);",
                optional.type, optional.example(0), convention.get("item"))
            .build())
        .runTest();
  }

  @Test
  public void testUpgradeMishapCaughtInCompiler() {
    // Users upgrading from older FreeBuilder versions may have implemented their own primitive
    // setter delegating to the OptionalP-accepting setter. We need to turn this runtime stack
    // overflow into a compile-time error.
    behaviorTester
        .with(new Processor(features))
        .with(SourceBuilder.forTesting(features)
            .addLine("package com.example;")
            .addLine("@%s", FreeBuilder.class)
            .addLine("@%s(builder = DataType.Builder.class)", JsonDeserialize.class)
            .addLine("public interface DataType {")
            .addLine("  %s %s;", optional.type, convention.get("item"))
            .addLine("")
            .addLine("  class Builder extends DataType_Builder {")
            .addLine("    public Builder %s(%s item) {",
                convention.set("item"), optional.primitiveType)
            .addLine("      return %s(%s.of(item));", convention.set("item"), optional.type)
            .addLine("    }")
            .addLine("  }")
            .addLine("")
            .addLine("  public static Builder builder() {")
            .addLine("    return new Builder();")
            .addLine("  }")
            .addLine("}"))
        .failsToCompile()
        .withErrorThat(error -> error
            .hasMessage("Infinite recursive loop detected")
            .inFile("/com/example/DataType.java")
            .onLine(14));
  }

  private static TestBuilder testBuilder() {
    return new TestBuilder()
        .addImport("com.example.DataType");
  }
}
